import os.path
from dataclasses import dataclass
import logging
from typing import Optional

import requests
from promptops import settings
from promptops.similarity import VectorDB
from promptops.trace import trace_id
from promptops.user import user_id
from promptops import scrub_secrets


@dataclass
class QATuple:
    question: str
    answer: str
    corrected: str

    @staticmethod
    def from_dict(data: dict) -> "QATuple":
        return QATuple(
            question=data["question"],
            answer=data["answer"],
            corrected=data["corrected"],
        )

    def to_dict(self):
        return {
            "question": self.question,
            "answer": self.answer,
            "corrected": self.corrected,
        }


_db: Optional[VectorDB] = None


def get_db() -> VectorDB:
    global _db
    if _db is None:
        _db = VectorDB()
        path = os.path.expanduser(settings.corrections_db_path)
        if os.path.exists(path):
            logging.debug("loading corrections db")
            _db.load(path)
        logging.debug(f"loaded corrections db: {len(_db)} records")

    return _db


def get_correction(prompt: str, command: str, error: str) -> Optional[str]:
    resp = requests.post(
        settings.endpoint + "/correction",
        json={
            "prompt": prompt,
            "command": scrub_secrets.scrub_line(".bash_history", command),
            "error": error,
            "model": settings.model,
            "trace_id": trace_id,
        },
        headers={
            "user-agent": f"promptops-cli; user_id={user_id()}",
        },
    )

    try:
        resp.raise_for_status()
    except Exception as e:
        logging.debug(f"error getting correction: {e}, {resp.text}")
        raise
    # best effort for now
    result = resp.json().get("command", None)

    return result



from argparse import ArgumentParser, REMAINDER
import sys
import os
import traceback

import logging

from promptops import settings
from promptops import settings_store
from promptops import version_check
from promptops import query
from promptops import user
from promptops.recipes.creation import workflow_entrypoint

ENDPOINT_ENV = "PROMPTOPS_ENDPOINT"


def runner_mode(args):
    from feedback import feedback

    feedback(
        {
            "event": "runner_mode",
        }
    )

    from local_runner.main import entry_point

    entry_point()


def handle_exception(prev_handler):
    def inner(exc_type, exc_value, exc_traceback):
        sys.stderr.flush()
        if prev_handler:
            prev_handler(exc_type, exc_value, exc_traceback)
        error_message = traceback.format_exception(exc_type, exc_value, exc_traceback)
        from promptops.feedback import feedback

        feedback(
            {
                "event": "unhandled_exception",
                "error": error_message,
            }
        )

    return inner


def entry_alias():
    # Set the global exception handler
    sys.excepthook = handle_exception(sys.excepthook)

    import colorama

    colorama.init()
    # get the name of the calling script
    alias = os.path.basename(sys.argv[0])

    settings_store.load()
    if endpoint := os.environ.get(ENDPOINT_ENV):
        settings.endpoint = endpoint
    parser = ArgumentParser(
        prog=alias,
        usage=f"{alias} [options] <question>\nexample: {alias} list running ec2 instances",
        description=f"{alias}: a command line assistant",
    )
    parser.add_argument("--verbose", "-v", action="store_true", help="verbose output")
    parser.add_argument("--version", action="store_true", help="print version and exit")
    parser.add_argument("--config", action="store_true", help="reconfigure")
    parser.add_argument(
        "--history-context",
        default=settings.history_context,
        type=int,
        help="past commands to include in the query (default: %(default)s)",
    )
    parser.add_argument(
        "--explain",
        dest="explain",
        action="store_true",
        default=settings.request_explanation,
        help="explain the commands and the parameters",
    )
    parser.add_argument(
        "--no-explain",
        dest="explain",
        action="store_false",
        default=settings.request_explanation,
        help="no explanations = faster response",
    )
    parser.add_argument(
        "--mode", default=settings.model, choices=["fast", "accurate"], help="fast or accurate (default: %(default)s)"
    )
    parser.add_argument("question", nargs=REMAINDER, help="the question to ask")
    registered = user.has_registered()
    args = parser.parse_args()
    if args.version:
        from promptops.version import __version__

        print(__version__)
        r = version_check.version_check()
        if not r.update_required:
            print("latest version:", r.latest_version)
        sys.exit(0)

    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO, format="%(message)s")
    version_check.version_check()
    if not registered or args.config:
        user.register()
    else:
        from promptops import history

        history.update_history()
    query.query_mode(args)


def entry_main():
    # Set the global exception handler
    sys.excepthook = handle_exception(sys.excepthook)

    import colorama

    colorama.init()

    settings_store.load()
    if endpoint := os.environ.get(ENDPOINT_ENV):
        settings.endpoint = endpoint
    parser = ArgumentParser()
    parser.add_argument("--version", action="store_true", help="print version and exit")
    parser.add_argument("--verbose", "-v", action="store_true", help="verbose output")
    parser.add_argument("--config", action="store_true", help="reconfigure")
    subparsers = parser.add_subparsers()
    parser_question = subparsers.add_parser("query", help="ask questions")
    parser_question.add_argument(
        "--history-context",
        default=settings.history_context,
        type=int,
        help="number of past commands to include in the query (default: %(default)s)",
    )
    parser_question.add_argument(
        "--explain",
        dest="explain",
        action="store_true",
        default=settings.request_explanation,
        help="explain the commands and the parameters",
    )
    parser_question.add_argument(
        "--no-explain",
        dest="explain",
        action="store_false",
        default=settings.request_explanation,
        help="get faster response",
    )
    parser_question.add_argument(
        "--mode", default=settings.model, choices=["fast", "accurate"], help="fast or accurate (default: %(default)s)"
    )
    parser_question.add_argument("question", nargs=REMAINDER, help="the question to ask")
    parser_question.set_defaults(func=query.query_mode)

    parser_runner = subparsers.add_parser("runner", help="run commands from slack")
    parser_runner.set_defaults(func=runner_mode)

    parser_workflow = subparsers.add_parser("workflow", help="run a complex or multi-stepped script")
    parser_workflow.add_argument("question", nargs=REMAINDER, help="the question to generate scripts for")
    parser_workflow.set_defaults(func=workflow_entrypoint)

    args = parser.parse_args()

    registered = user.has_registered()

    if not hasattr(args, "func"):
        if args.version:
            from promptops.version import __version__

            print(__version__)

            r = version_check.version_check()
            if not r.update_required:
                print("latest version:", r.latest_version)
            sys.exit(0)
        if args.config:
            user.register()
            sys.exit(0)

        parser.print_help()
        return

    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO, format="%(message)s")

    version_check.version_check()
    if not registered:
        user.register()
    else:
        from promptops import history

        history.update_history()
    args.func(args)


if __name__ == "__main__":
    entry_main()



from promptops.shells.base import Shell, reverse_readline, accept_command
from promptops.scrub_secrets import scrub_lines
import os


class Bash(Shell):
    def __init__(self, history_file: str = None):
        if not history_file:
            history_file = os.getenv("HISTFILE", "~/.bash_history")
        super().__init__(history_file)

    def get_recent_history(self, look_back: int = 10):
        fname = os.path.expanduser(self.history_file)
        buffer = ""
        commands = []
        for line in reverse_readline(fname):
            if len(commands) >= look_back:
                break
            current = line.rstrip()
            if current.endswith("\\"):
                buffer = current + "\n" + buffer
                continue
            else:
                if accept_command(buffer):
                    commands.append(buffer)
                buffer = current
        return scrub_lines(fname, list(reversed(commands)))

    def _get_cmds_from_lines(self, lines):
        buffer = ""
        commands = []
        for line in lines:
            buffer += "\n" + line.rstrip()
            if not line.endswith("\\"):
                commands.append(buffer.lstrip())
                buffer = ""
        return commands



DEFAULT_ENDPOINT = "https://cli.promptops.com"


endpoint: str = DEFAULT_ENDPOINT
request_explanation: bool = True
model: str = "accurate"
history_context: int = 0
corrections_db_path = "~/.promptops/corrections.db"
history_db_path = "~/.promptops/history.db"
user_id_path = "~/.promptops/user_id"
index_history: bool = False



import re

from promptops.shells.base import Shell, accept_command, reverse_readline
from promptops.scrub_secrets import scrub_lines
import os


def _is_zsh_start_line(line):
    return re.match(r"^: \d+:\d+;.+", line) is not None


class Zsh(Shell):
    def __init__(self, history_file: str = None):
        if not history_file:
            history_file = os.getenv("HISTFILE", "~/.zsh_history")
        super().__init__(history_file)

    def get_recent_history(self, look_back: int = 10):
        fname = os.path.expanduser(self.history_file)
        buffer = ""
        commands = []
        for line in reverse_readline(fname):
            if len(commands) >= look_back:
                break
            if _is_zsh_start_line(line):
                line = line.split(";")[1].rstrip()
                if line.endswith("\\"):
                    line = line[:-1]
                cmd = line + buffer
                buffer = ""
                if accept_command(cmd):
                    commands.append(cmd)
            else:
                line = line.rstrip()
                if line.endswith("\\"):
                    line = line[:-1]
                buffer = "\n" + line + buffer
        return scrub_lines(fname, list(reversed(commands)))

    def _get_cmds_from_lines(self, lines):
        buffer = ""
        commands = []
        for line in lines:
            if _is_zsh_start_line(line):
                if buffer != "":
                    commands.append(buffer)
                buffer = line.split(";")[1].rstrip()
            else:
                buffer += "\n" + line.rstrip()
            if buffer.endswith("\\"):
                buffer = buffer[:-1]
        if buffer != "" and not buffer.endswith("\\"):
            commands.append(buffer)
        return commands



import threading
from contextlib import contextmanager


def animate(loader, event: threading.Event, fps=10):
    while not event.is_set():
        loader.step()
        event.wait(1 / fps)
    loader.clear()


@contextmanager
def loading_animation(loader, fps=10):
    event = threading.Event()
    t = threading.Thread(target=animate, args=(loader, event), kwargs=dict(fps=fps), daemon=True)
    t.start()
    try:
        yield
    finally:
        event.set()
        t.join()



import asyncio
from asyncio import Queue
from dataclasses import dataclass
from typing import Callable, List, Optional
import subprocess


@dataclass
class Result:
    exit_code: Optional[int] = None
    error: Optional[Exception] = None


@dataclass
class Request:
    cmd: str
    args: List[str]
    on_started: Callable
    on_completed: Callable


@dataclass
class Status:
    queued: int
    running: int
    completed: int


class Manager:
    def __init__(self, parallelism: int = 1):
        self._queue = Queue()
        self.parallelism = max(parallelism, 1)
        self.busy = 0
        self.completed = 0
        self.worker_tasks = []

    def start(self):
        for _ in range(self.parallelism):
            worker = asyncio.create_task(self.worker())
            self.worker_tasks.append(worker)

    async def worker(self):
        while True:
            request = await self._queue.get()
            if request is None:
                # shutdown
                return
            self.busy += 1
            try:
                await self.run(request)
            except Exception as e:
                request.on_completed(Result(exit_code=None, error=e))
            finally:
                # just in case we decide to call join() on the queue
                self._queue.task_done()
                self.busy -= 1
                self.completed += 1

    async def queue(self, request: Request):
        await self._queue.put(request)

    def status(self) -> Status:
        return Status(queued=self._queue.qsize(), running=self.busy, completed=self.completed)

    @staticmethod
    async def run(request: Request):
        process = await asyncio.create_subprocess_exec(
            request.cmd,
            *request.args,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE
        )

        request.on_started(process.stdout, process.stderr)

        await process.wait()
        request.on_completed(Result(exit_code=process.returncode))

    async def stop(self, timeout: Optional[float] = None):
        for _ in self.worker_tasks:
            await self._queue.put(None)

        if timeout is not None:
            await asyncio.wait_for(asyncio.gather(*self.worker_tasks), timeout=timeout)
        else:
            await asyncio.gather(*self.worker_tasks)



from detect_secrets import SecretsCollection
from detect_secrets.settings import default_settings


def scrub_file(filename, lines):
    secrets = SecretsCollection()
    lines_to_secrets = {}

    with default_settings():
        secrets.scan_file(filename)
        for secret in secrets.data[filename]:
            if lines_to_secrets.get(secret.line_number - 1) is None:
                lines_to_secrets[secret.line_number - 1] = [secret]
            else:
                lines_to_secrets[secret.line_number - 1].append(secret)

    for index, secrets in lines_to_secrets.items():
        for secret in secrets:
            lines[index] = lines[index].replace(secret.secret_value, "<SECRET>")

    return lines


def scan_lines(fake_filename: str, lines: list[str]):
    from detect_secrets.settings import get_plugins
    from detect_secrets.core.scan import _scan_line, _is_filtered_out
    from detect_secrets.util.code_snippet import get_code_snippet

    for line_number, line in enumerate(lines, start=1):
        code_snippet = get_code_snippet(lines, line_number)

        yield from (
            secret
            for plugin in get_plugins()
            for secret in _scan_line(
                plugin=plugin,
                filename=fake_filename,
                line=line,
                line_number=line_number,
                context=code_snippet,
            )
            if not _is_filtered_out(
                required_filter_parameters=["context"],
                filename=secret.filename,
                secret=secret.secret_value,
                plugin=plugin,
                line=line,
                context=code_snippet,
            )
        )


_PADDING = [
    "echo hello world",
    "cat test > test.txt",
    'git commit -m "um: padding for secrets"',
    "git pull --rebase",
    "git push",
]


def scrub_lines(fake_filename: str, lines: list[str]) -> list[str]:
    # copy the lines, we'll update in place
    scrubbed = [line for line in lines]
    with default_settings():
        for secret in scan_lines(fake_filename, _PADDING + lines + _PADDING):
            i = secret.line_number - 1 - len(_PADDING)
            if i < 0 or i > len(lines):
                continue
            scrubbed[i] = lines[i].replace(secret.secret_value, "<SECRET>")
    return scrubbed


def scrub_line(fake_filename: str, line: str) -> str:
    return scrub_lines(fake_filename, [line])[0]



import requests
from .dtos import Request, HeartbeatResponse, UpdateMessage, Result
from .base import Reporter


class HttpReporter(Reporter):
    def __init__(self, session: requests.Session):
        self.session = session

    @staticmethod
    def _make_headers(request: Request):
        return {
            "Authorization": f"Bearer {request.token}",
            "Content-Type": "application/json"
        }

    def heartbeat(self, request: Request) -> HeartbeatResponse:
        path = f"/api/v1/session/{request.session_id}/heartbeat"
        url = request.state_service_url + path
        body = {
            "blockIndex": request.block_index
        }
        response = self.session.post(url, json=body, headers=self._make_headers(request))
        response.raise_for_status()
        return HeartbeatResponse.from_dict(response.json())

    def report_result(self, request: Request, result: Result) -> None:
        path = f"/api/v1/session/{request.session_id}/result"
        url = request.state_service_url + path
        body = result.to_dict()
        response = self.session.post(url, json=body, headers=self._make_headers(request))
        response.raise_for_status()

    def report_update(self, request: Request, update: UpdateMessage) -> None:
        path = f"/api/v1/session/{request.session_id}/update"
        url = request.state_service_url + path
        body = update.to_dict()
        response = self.session.post(url, json=body, headers=self._make_headers(request))
        response.raise_for_status()



from dataclasses import dataclass
from typing import Any


@dataclass
class Code:
    type: str = "python"
    content: str = ""

    def to_dict(self):
        return {
            "type": self.type,
            "content": self.content,
        }

    @staticmethod
    def from_dict(d: dict) -> "Code":
        return Code(
            type=d.get("type", ""),
            content=d.get("content", ""),
        )


@dataclass
class Request:
    token: str
    session_id: str
    block_index: int
    code: Code
    state_service_url: str

    def to_dict(self):
        return {
            "token": self.token,
            "sessionId": self.session_id,
            "blockIndex": self.block_index,
            "code": self.code.to_dict(),
            "stateServiceUrl": self.state_service_url,
        }

    @staticmethod
    def from_dict(d: dict) -> "Request":
        return Request(
            token=d.get("token", ""),
            session_id=d.get("sessionId", ""),
            block_index=d.get("blockIndex", 0),
            code=Code.from_dict(d.get("code", {})),
            state_service_url=d.get("stateServiceUrl", ""),
        )


@dataclass
class Message:
    type: str = ""
    payload: Any = None

    def to_dict(self):
        return {
            "type": self.type,
            "payload": self.payload,
        }

    @staticmethod
    def from_dict(d: dict) -> "Message":
        return Message(
            type=d.get("type", ""),
            payload=d.get("payload"),
        )


@dataclass
class HeartbeatResponse:
    signal: str
    metadata: Any

    def to_dict(self):
        return {
            "signal": self.signal,
            "metadata": self.metadata,
        }

    @staticmethod
    def from_dict(d: dict) -> "HeartbeatResponse":
        return HeartbeatResponse(
            signal=d.get("signal", ""),
            metadata=d.get("metadata", ""),
        )


@dataclass
class UpdateMessage:
    message_type: str
    block_index: int
    data: "StreamData"
    time: int

    def to_dict(self):
        return {
            "messageType": self.message_type,
            "blockIndex": self.block_index,
            "data": self.data.to_dict(),
            "time": self.time,
        }


@dataclass
class Result:
    status: int
    reason: str
    result: dict
    block_index: int

    def to_dict(self):
        return {
            "status": self.status,
            "reason": self.reason,
            "result": self.result,
            "blockIndex": self.block_index,
        }


@dataclass
class StreamData:
    stream: str
    data: str

    def to_dict(self):
        return {
            "stream": self.stream,
            "data": self.data,
        }


@dataclass
class StructuredOutput:
    data: str
    encoding: str

    def to_dict(self):
        return {
            "data": self.data,
            "encoding": self.encoding,
        }



import json
import websockets
from typing import Callable
import logging


class Listener:
    def __init__(self, handler: Callable):
        self.handler = handler

    async def listen(self, url: str, get_token: Callable):
        async with websockets.connect(url) as websocket:
            auth_needed = True
            while True:
                if auth_needed:
                    token = get_token()
                    await websocket.send(json.dumps({
                        "__action__": "authorize",
                        "token": token
                    }))
                message = await websocket.recv()
                data = json.loads(message)

                action = data.get("__action__")
                if action:
                    if action == "authorization_success":
                        print("Ready")
                        auth_needed = False
                    elif action in ["no_authorization", "authorization_failure"]:
                        raise ValueError(f"{action}: {data.get('__text__')}")
                    else:
                        raise ValueError(f"Unknown action: {action}")
                else:
                    try:
                        await self.handler(data, websocket)
                    except Exception as e:
                        logging.error("error handling message", exc_info=e)



from .base import Shell
from .base import reverse_readline, accept_command
from promptops.scrub_secrets import scrub_lines
import os


def _is_start_line(line):
    return line.startswith("- cmd: ")


class Fish(Shell):
    def __init__(self, history_file: str = None):
        if not history_file:
            history_file = os.getenv("HISTFILE", "~/.local/share/fish/fish_history")
            if not os.path.exists(os.path.expanduser(history_file)):
                history_file = "~/.config/fish/fish_history"

        super().__init__(history_file)

    def _get_cmds_from_lines(self, lines):
        commands = []
        for line in lines:
            if _is_start_line(line):
                cmd = line.split("- cmd: ")[1].rstrip()
                # unescape see https://stackoverflow.com/a/57192592
                cmd = cmd.encode("latin-1", "backslashreplace").decode("unicode-escape")
                if accept_command(cmd):
                    commands.append(cmd)
        return commands

    def get_recent_history(self, look_back: int = 10):
        fname = os.path.expanduser(self.history_file)
        commands = []
        for line in reverse_readline(fname):
            if len(commands) >= look_back:
                break
            if _is_start_line(line):
                cmd = line.split("- cmd: ")[1].rstrip()
                # unescape see https://stackoverflow.com/a/57192592
                cmd = cmd.encode("latin-1", "backslashreplace").decode("unicode-escape")
                if accept_command(cmd):
                    commands.append(cmd)
        return scrub_lines(fname, list(reversed(commands)))



QUERY = [
    "hold on, fetching data from the Matrix...",
    "let me consult the command oracle...",
    "gathering ancient shell wisdom...",
    "contacting the shell gurus on the other side...",
    "talking to the shell spirits, one moment please...",
    "scouring the shell dungeons for answers...",
    "scanning the shell multiverse...",
    "whispering to the shell command fairies...",
    "asking the shell command crystal ball...",
    "brewing a potion to summon the right command...",
    "reading the command tarot cards...",
    "channeling the command force, stay tuned...",
]

EXPLAIN = [
    "let me decrypt this shell spell for you...",
    "translating command hieroglyphs...",
    "hold on, I'm unlocking the shell command secrets...",
    "it's shell command storytime, one moment...",
    "digging up the hidden meaning of this command...",
    "shell command mysteries, being revealed shortly...",
    "peeling back the layers of this shell command onion...",
    "decoding the command enigma...",
    "cracking open the command vault of knowledge...",
    "navigating the shell command labyrinth...",
    "solving the command riddle for you...",
    "assembling the command puzzle pieces...",
]



from dataclasses import dataclass


@dataclass
class Result:
    script: str
    explanation: str = None
    lang: str = "shell"
    origin: str = "promptops"

    @staticmethod
    def from_dict(data: dict) -> "Result":
        return Result(
            script=data["script"],
            explanation=data.get("explanation"),
            lang=data.get("lang", "shell"),
            origin=data.get("origin", "promptops"),
        )



import abc
from .dtos import Request, HeartbeatResponse, Result, UpdateMessage


class Reporter(abc.ABC):
    @abc.abstractmethod
    def heartbeat(self, request: Request) -> HeartbeatResponse:
        pass

    @abc.abstractmethod
    def report_result(self, request: Request, result: Result) -> None:
        pass

    @abc.abstractmethod
    def report_update(self, request: Request, update: UpdateMessage) -> None:
        pass



import sys
import time
import threading
import os
import typing
from promptops.loading.simple import loader
from prompt_toolkit.formatted_text import to_plain_text, ANSI
import wcwidth
import colorama


def getch():
    if sys.platform == "win32":
        import msvcrt

        return msvcrt.getwch()
    else:
        import tty
        import termios

        fd = sys.stdin.fileno()
        old = termios.tcgetattr(fd)
        try:
            tty.setraw(fd)
            return sys.stdin.read(1)
        finally:
            termios.tcsetattr(fd, termios.TCSADRAIN, old)


FOOTER_SECTIONS = {
    "select": f"{colorama.Style.BRIGHT}[↑/↓]{colorama.Style.RESET_ALL} select",
    "confirm": f"{colorama.Style.BRIGHT}[enter]{colorama.Style.RESET_ALL} confirm",
    "cancel": f"{colorama.Style.BRIGHT}[ctrl+c]{colorama.Style.RESET_ALL} cancel",
    "copy": f"{colorama.Style.BRIGHT}[c]{colorama.Style.RESET_ALL} copy",
}


class UI(object):
    def __init__(
        self,
        options: list[str],
        is_loading: bool,
        cursor="➜︎",
        show_timer=False,
        loading_text="getting you the best results ...",
        footer=" ".join([FOOTER_SECTIONS["select"], FOOTER_SECTIONS["confirm"], FOOTER_SECTIONS["cancel"]]),
        actions: dict[str, typing.Callable[[str, "UI"], None]] = None,
    ):
        self._selected = 0
        self._options = options
        self._is_loading = is_loading
        self._start = time.time()
        self._cursor = cursor
        self._footer = footer
        self._loading_line_index = None
        self._option_indexes = []
        self._thread = threading.Thread(target=self._loading_animation, daemon=True)
        self._show_timer = show_timer
        self._spinner = loader(loading_text)
        self._lock = threading.Lock()
        self._is_active = True
        self._actions = actions or {}

        self._lines_written = 0
        self.render()
        self._thread.start()

    def render(self):
        with self._lock:
            # clean all the lines that were written
            sys.stdout.write("\r\x1b[K")
            for _ in range(self._lines_written):
                sys.stdout.write("\x1b[1A\x1b[K")
            size = os.get_terminal_size()
            width = size.columns
            self._option_indexes = []

            current_line = 0
            for i, option in enumerate(self._options):
                self._option_indexes.append(current_line)
                if i == self._selected:
                    text = f" {self._cursor} {self._get_formatted_text(option, True)}"
                else:
                    text = f"   {self._get_formatted_text(option, False)}"
                sys.stdout.write(f"\r{text}")
                sys.stdout.write("\n\r")
                for line in to_plain_text(ANSI(text)).split("\n"):
                    current_line += 1
                    text_width = wcwidth.wcswidth(line)
                    if text_width > width:
                        current_line += text_width // width

            if self._is_loading:
                self._loading_line_index = current_line
                if self._show_timer:
                    sys.stdout.write(f"\r  {self._spinner(advance=False)} {(time.time() - self._start):.2f}s\n\r")
                else:
                    sys.stdout.write(f"\r  {self._spinner(advance=False)}\n\r")
                current_line += 1

            if self._footer:
                sys.stdout.write(f"\n\r{self._footer}")
                current_line += 1

            self._lines_written = current_line

            sys.stdout.flush()

    def _loading_animation(self):
        while True:
            with self._lock:
                if self._is_loading:
                    # sys.stdout.write("\x1b[s")
                    sys.stdout.write("\x1b7")
                    sys.stdout.write(f"\x1b[{self._lines_written - self._loading_line_index}A\x1b[K")
                    if self._show_timer:
                        sys.stdout.write(f"\r   {self._spinner()} {(time.time() - self._start):.2f}s\n\r")
                    else:
                        sys.stdout.write(f"\r   {self._spinner()}\n\r")
                    # sys.stdout.write("\x1b[u")
                    sys.stdout.write("\x1b8")
                    sys.stdout.flush()

            time.sleep(0.1)

    def select(self, index):
        if index == self._selected:
            return
        if len(self._options) == 0:
            return
        with self._lock:
            sys.stdout.write("\x1b7")
            sys.stdout.write(f"\x1b[{self._lines_written - self._option_indexes[index]}A")
            sys.stdout.write(f"\r {self._cursor} {self._get_formatted_text(self._options[index], True)}")
            sys.stdout.write("\x1b8")
            sys.stdout.write("\x1b7")
            sys.stdout.write(f"\x1b[{self._lines_written - self._option_indexes[self._selected]}A")
            sys.stdout.write(f"\r\x1b[K   {self._get_formatted_text(self._options[self._selected], False)}")
            sys.stdout.write(f"\x1b[{self._lines_written}B\r")
            sys.stdout.write("\x1b8")
            sys.stdout.flush()
            self._selected = index

    @property
    def selected(self):
        return self._selected

    @property
    def options(self):
        return self._options

    @property
    def is_loading(self):
        return self._is_loading

    def _get_formatted_text(self, option: str, selected: bool):
        return f"{colorama.Style.BRIGHT if selected else ''}{option}{colorama.Style.RESET_ALL}"

    def add_options(self, options, is_loading=False):
        if not self._is_active:
            return
        self._options.extend(options)
        self._is_loading = is_loading
        self.render()

    def reset_options(self, options, is_loading=False):
        if not self._is_active:
            return
        self._options = options
        if self._selected >= len(self._options):
            self._selected = len(self._options) - 1
        self._is_loading = is_loading
        self.render()

    def input(self):
        while True:
            key = getch()
            if key == "\x03":
                if self._footer:
                    sys.stdout.write("\r\x1b[K")
                raise KeyboardInterrupt()
            elif key == "\x1b":
                next_ch = getch()
                if next_ch == "[":
                    last_ch = getch()
                    if last_ch == "A":
                        index = max(0, self._selected - 1)
                        self.select(index)
                    elif last_ch == "B":
                        index = min(len(self._options) - 1, self._selected + 1)
                        self.select(index)
            elif key in ["\r", "\n"]:
                self._is_active = False
                self._is_loading = False
                self._footer = None
                self.render()
                return self._selected
            elif key in self._actions:
                self._actions[key](key, self)



import requests
from promptops import settings
from promptops import user
from promptops import version
from prompt_toolkit import print_formatted_text
from prompt_toolkit.formatted_text import HTML
from dataclasses import dataclass
from promptops.trace import trace_id


@dataclass
class Response:
    update_required: bool = False
    message: str = None
    latest_version: str = None

    @staticmethod
    def from_dict(data: dict) -> "Response":
        return Response(
            update_required=data.get("update_required", False),
            message=data.get("message"),
            latest_version=data.get("latest_version"),
        )


def version_check():
    response = requests.post(
        settings.endpoint + "/version",
        json={
            "version": version.__version__,
            "trace_id": trace_id,
        },
        headers={"user-agent": f"promptops-cli; user_id={user.user_id()}"},
    )
    if response.status_code != 200:
        print("version check failed", response.status_code, response.text)
    data = response.json()
    r = Response.from_dict(data)
    if r.update_required:
        print()
        print_formatted_text(HTML("    <bold>update is required</bold>"))
        print_formatted_text(
            HTML(
                f"    current version: <ansired>{version.__version__}</ansired>, latest: <ansigreen>{r.latest_version}</ansigreen>"
            )
        )
        message = "to update run <ansigreen>pip install --upgrade git+https://github.com/promptops/cli.git</ansigreen>"
        print_formatted_text(HTML(f"    {r.message or message}"))
        print()
    return r



from promptops import trace
from promptops import settings
from promptops import user
import requests


def feedback(payload: dict):
    requests.post(settings.endpoint + "/feedback", json={
        "trace_id": trace.trace_id,
        **payload
    }, headers={
        "user-agent": f"promptops-cli; user_id={user.user_id()}"
    })
    # feedback is best-effort, so we don't raise for status



import asyncio

import requests

from local_runner.comms import Listener, HttpReporter
from local_runner import config
from local_runner.tokens import Issuer
from local_runner import creds
from local_runner import exec
from local_runner.registration import register
from local_runner import handler


async def _main():
    cfg = config.default()
    try:
        c = creds.default(cfg.credentials_file)
    except creds.NotFoundError:
        print("Credentials not found. Initiating pairing...")
        c = await asyncio.wait_for(register(cfg.registration_url, cfg.registration_code_rotation), cfg.registration_timeout)
        creds.store(cfg.credentials_file, c)

    issuer = Issuer(c)
    mgr = exec.Manager(2)
    mgr.start()

    session = requests.Session()
    reporter = HttpReporter(session)
    h = handler.Handler(mgr, reporter, cfg)
    listener = Listener(h.handle)
    print("[ctrl+c] to exit")
    try:
        await listener.listen(cfg.ws_url, issuer.get_token)
    except asyncio.CancelledError:
        print("shutting down...")
    await mgr.stop(10.0)


def entry_point():
    try:
        asyncio.run(_main())
    except KeyboardInterrupt:
        print("bye")


if __name__ == "__main__":
    entry_point()



import os
from dataclasses import dataclass


@dataclass
class Config:
    ws_url: str
    auth_url: str
    credentials_file: str
    python_path: str
    shell_path: str
    registration_url: str
    registration_timeout: int
    registration_code_rotation: int
    parallelism: int


ENV_PREFIX = "PROMPTOPS"

ENV_WS_URL = f"{ENV_PREFIX}_WS_URL"
ENV_AUTH_URL = f"{ENV_PREFIX}_AUTH_URL"
ENV_CREDENTIALS_FILE = f"{ENV_PREFIX}_CREDENTIALS_FILE"
ENV_PYTHON_PATH = f"{ENV_PREFIX}_PYTHON_PATH"
ENV_SHELL_PATH = f"{ENV_PREFIX}_SHELL_PATH"
ENV_REGISTRATION_URL = f"{ENV_PREFIX}_REGISTRATION_URL"


def default() -> Config:
    cfg = Config(
        ws_url="wss://local-runner-api.global.ctrlstack.com",
        auth_url="https://authorization-tokens.global.ctrlstack.com/token",
        credentials_file="~/.promptops/credentials.json",
        python_path="python",
        shell_path=os.getenv("SHELL"),
        registration_url="wss://local-runner-registrar.global.ctrlstack.com",
        registration_timeout=30 * 60,  # 30 minutes in seconds
        registration_code_rotation=5 * 60,  # 5 minutes in seconds
        parallelism=2,  # the number of concurrent tasks to run
    )

    if (url := os.getenv(ENV_WS_URL)) is not None:
        cfg.ws_url = url
    if (url := os.getenv(ENV_AUTH_URL)) is not None:
        cfg.auth_url = url
    if (file := os.getenv(ENV_CREDENTIALS_FILE)) is not None:
        cfg.credentials_file = file
    if (path := os.getenv(ENV_PYTHON_PATH)) is not None:
        cfg.python_path = path
    if (path := os.getenv(ENV_SHELL_PATH)) is not None:
        cfg.shell_path = path
    if (url := os.getenv(ENV_REGISTRATION_URL)) is not None:
        cfg.registration_url = url

    return cfg



from promptops.shells.zsh import Zsh
from promptops.shells.bash import Bash
from promptops.shells.fish import Fish
from promptops.shells.base import NoopShell
import os
import psutil
import logging


_shells = {
    "zsh": Zsh,
    "bash": Bash,
    "fish": Fish,
}


def _resolve_shell():
    proc = psutil.Process(os.getpid())

    path = []
    while proc is not None and proc.pid > 0:
        try:
            name = proc.name()
        except TypeError:
            name = proc.name

        name = os.path.splitext(name)[0]
        path = [name] + path

        if name in _shells:
            logging.debug(f"found shell: {'/'.join(path)}")
            return _shells[name]()

        try:
            proc = proc.parent()
        except TypeError:
            proc = proc.parent

    from promptops.feedback import feedback

    feedback(
        {
            "event": "shell_not_found",
            "path": "/".join(path),
        }
    )
    logging.debug(f"shell not supported: {'/'.join(path)}")
    return NoopShell()


def get_shell():
    shell = _resolve_shell()
    return shell



import logging
import os
import sys
from dataclasses import dataclass
from copy import copy
import subprocess
import threading
import random
from typing import Optional

import requests
from prompt_toolkit import print_formatted_text
from prompt_toolkit.patch_stdout import patch_stdout
from prompt_toolkit.formatted_text import HTML, to_formatted_text
import prompt_toolkit

from promptops import settings
from promptops import trace
from promptops import user
from promptops.ui import prompts
from promptops.ui import selections
from promptops.cancellable import ThreadPoolExecutor
from promptops.loading import Simple, loading_animation
from promptops import similarity
from promptops import corrections
from promptops.corrections import get_correction
from promptops.feedback import feedback
from promptops import history
from promptops import settings_store
from promptops import scrub_secrets
from .dtos import Result
from .explanation import get_explanation, ReturningThread
from . import messages


def deduplicate(results: list[Result]):
    results_set = set()
    final_results = []
    for result in results:
        if result.script in results_set:
            continue
        results_set.add(result.script)
        final_results.append(result)
    return final_results


def run(cmd: Result) -> (int, Optional[str]):
    if cmd.lang == "shell":
        proc = subprocess.run(
            cmd.script, shell=True, start_new_session=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        if proc.stdout and str(proc.stdout) != "":
            sys.stdout.write(proc.stdout.decode("utf-8"))
        if proc.stderr and str(proc.stderr) != "":
            sys.stdout.write(proc.stderr.decode("utf-8"))
        sys.stdout.flush()

        history.add(scrub_secrets.scrub_line(".bash_history", cmd.script), proc.returncode)
        return proc.returncode, str(proc.stderr)
    else:
        raise NotImplementedError(f"{cmd.lang} not implemented yet")


def corrections_search(embedding):
    db = corrections.get_db()
    similar = db.search(embedding, k=3, min_similarity=0.8)
    return [(corrections.QATuple.from_dict(s), score) for s, score in similar]


def make_revise_option():
    return "\x1b[3m💭️ don't see what you're looking for? try providing more context\x1b[0m"


ORIGIN_SYMBOLS = {"history": "📖", "promptops": "✨"}


def pretty_result(result: Result):
    return (ORIGIN_SYMBOLS.get(result.origin, f"[{result.origin[0]}]") + " " if result.origin else "") + result.script


@dataclass
class ConfirmResult:
    result: Result = None
    confirmed: str = ""
    question: str = None
    options: list = None


def revise_loop(questions: list[str], prev_results: list[list[str]]) -> ConfirmResult:
    embedding = similarity.embedding(text="\n".join(questions))
    with loading_animation(Simple("thinking...")), ThreadPoolExecutor(max_workers=5) as tpe:
        cs_task = tpe.submit(corrections_search, embedding)
        hs_task = tpe.submit(history.check_history, embedding)

    update_lock = threading.Lock()
    corrected_results = []
    similar_corrections = cs_task.result()
    logging.debug("found %d similar corrections", len(similar_corrections))
    if len(similar_corrections) > 0:
        for qa, score in similar_corrections:
            logging.debug(f"- {score:.2f}: {qa.question} -> {qa.corrected}")
            corrected_results.append(Result(script=qa.corrected, origin="history"))
    history_results = hs_task.result()
    logging.debug("found %d similar history results", len(history_results))
    if len(history_results) > 0:
        for r, score in history_results:
            logging.debug(f"- {score:.2f}: {r}")
    history_results = [
        Result(script=r if isinstance(r, str) else r["cmd"], explanation="loaded from shell history", origin="history")
        for r, _ in history_results
    ]

    results = corrected_results + history_results
    results = deduplicate(results)

    tpe = ThreadPoolExecutor(max_workers=2)
    ui = None
    tasks = [
        tpe.submit(
            query,
            q=questions[-1],
            prev_questions=questions[:-1],
            corrected_results=[qa for qa, _ in similar_corrections],
            prev_results=prev_results,
            similar_history=[r.script for r in history_results],
            history_context=[],
        )
    ]
    num_running = len(tasks)

    def update_results(extra):
        with update_lock:
            nonlocal num_running
            num_running -= 1
            # TODO: deduplicate
            results.extend(extra)
            if ui:
                options = [pretty_result(r) for r in results]
                if num_running == 0:
                    options += [make_revise_option()]
                ui.reset_options(options, is_loading=num_running > 0)

    def done_callback(future):
        if future.exception():
            print(future.exception())
        update_results(future.result())

    for task in tasks:
        task.add_done_callback(tpe.cancellable_callback(done_callback))

    if len(results) == 0:
        with loading_animation(Simple("thinking...")):
            tpe.shutdown(wait=True)
        num_running = 0
        results = [r for t in tasks for r in t.result()]
        results = deduplicate(results)
        if len(results) == 0:
            feedback({"event": "no-results"})
            print("couldn't find any results, try rephrasing your question or providing more context")
            selected = prompts.confirm_clarify("")
            if selected == prompts.EXIT:
                raise KeyboardInterrupt()
            return ConfirmResult(question=selected, options=[])

    def remove_entry(_, ui: selections.UI):
        with update_lock:
            if ui.selected >= len(results):
                return
            result = results[ui.selected]
            if result.origin == "history":
                results.pop(ui.selected)
                options = [pretty_result(r) for r in results]
                if num_running == 0:
                    options += [make_revise_option()]
                ui.reset_options(options, is_loading=num_running > 0)
        hdb = history.get_history_db()
        for item in hdb.objects:
            if isinstance(item, dict) and item["cmd"] == result.script:
                item["ignore"] = True
        hdb.save(os.path.expanduser(settings.history_db_path))

    while True:
        with update_lock:
            options = [pretty_result(r) for r in results]
            if num_running == 0:
                options += [make_revise_option()]
            ui = selections.UI(
                options,
                num_running > 0,
                loading_text=random.choice(messages.QUERY),
                actions={
                    "\x08": remove_entry,
                    "\x7F": remove_entry,
                },
            )
        index = ui.input()
        print()
        if index == len(results):
            selected = prompts.confirm_clarify("")
            if selected == prompts.GO_BACK:
                continue
            elif selected == prompts.EXIT:
                raise KeyboardInterrupt()
            return ConfirmResult(question=selected, options=[r.script for r in results if r.origin == "promptops"])
        else:
            selected = results[index]
            if selected.explanation:
                print_formatted_text(HTML(selected.explanation))
                print()
                confirmed = prompts.confirm_command(selected.script, False)
            elif selected.origin == "promptops" and settings.request_explanation:

                done_loading = threading.Event()

                def _explain_and_done(*args, **kwargs):
                    try:
                        return get_explanation(*args, **kwargs)
                    finally:
                        done_loading.set()

                def _confirm_and_done(*args, **kwargs):
                    try:
                        return prompts.confirm_command(*args, **kwargs)
                    finally:
                        done_loading.set()

                with patch_stdout(raw=True):
                    et = ReturningThread(
                        target=_explain_and_done,
                        kwargs=dict(
                            result=selected,
                            prev_questions=questions[:-1],
                            prev_results=prev_results,
                            corrected_results=[qa for qa, _ in similar_corrections],
                            similar_history=[r.script for r in history_results],
                            history_context=[],
                        ),
                        daemon=True,
                    )
                    pt = ReturningThread(
                        target=_confirm_and_done,
                        args=(selected.script, False),
                        kwargs=dict(
                            message="\n> ",
                        ),
                    )
                    msg = random.choice(messages.EXPLAIN)
                    sys.stdout.sleep_between_writes = 0.2
                    with loading_animation(Simple(msg), fps=5):
                        et.start()
                        pt.start()
                        done_loading.wait()
                    sys.stdout.sleep_between_writes = 0.5
                    if not et.is_alive():
                        try:
                            explanation = et.join()
                            indented = "\n".join(" │  " + line for line in explanation.splitlines())
                            print_formatted_text(to_formatted_text(HTML(indented), style="fg: ansiyellow"))
                            print()
                        except requests.HTTPError as e:
                            print_formatted_text(
                                HTML(f"<ansired> │  oh snap, failed to fetch you the explanation: {e}</ansired>")
                            )
                            print()
                    confirmed = pt.join()
            else:
                confirmed = prompts.confirm_command(selected.script, False)

            if confirmed == prompts.GO_BACK:
                continue
            elif confirmed == prompts.EXIT:
                raise KeyboardInterrupt()
            return ConfirmResult(result=selected, confirmed=confirmed)


def correction_loop(prompt: str, command: str, error: str) -> Optional[Result]:
    selected = prompts.confirm(
        f"looks like the command failed (return code was not 0), would you like us to attempt to fix it?"
    )
    if selected == prompts.GO_BACK:
        return None
    elif selected == prompts.EXIT:
        raise KeyboardInterrupt()

    with loading_animation(Simple("thinking...")), ThreadPoolExecutor(max_workers=5) as tpe:
        ce_task = tpe.submit(get_correction, prompt=prompt, command=command, error=error)

    result = ce_task.result()

    results = [
        Result(script=result, explanation=f"Revised previous command: {command}"),
    ]
    options = [pretty_result(r) for r in results]
    ui = selections.UI(options, False, loading_text=random.choice(messages.QUERY))

    index = ui.input()
    selected = results[index]
    if selected == prompts.GO_BACK:
        return None
    elif selected == prompts.EXIT:
        raise KeyboardInterrupt()

    print()
    return selected


def do_query(question: str):
    question = question.strip()
    if question == "":
        feedback({"event": "empty-initial-query"})
        for i in range(2):
            if i > 0:
                print("please enter a question")
            bottom_toolbar = HTML("<b>[enter]</b> confirm <b>[ctrl+c]</b> exit")
            # get the name of the calling script
            alias = os.path.basename(sys.argv[0])
            question = prompt_toolkit.prompt(f"{alias}: ", bottom_toolbar=bottom_toolbar)
            question = question.strip()
            if question != "":
                break
        else:
            print("no question entered")
            sys.exit(0)

    questions = [question]
    prev_results = []

    print()

    def input_loop():
        while True:
            try:
                results = revise_loop(questions, prev_results)
                if results.result:
                    return results.result, results.confirmed
                else:
                    print()
                    feedback({"event": "revise"})
                    questions.append(results.question)
                    prev_results.append(results.options)
            except KeyboardInterrupt:
                print()
                feedback({"event": "cancelled"})
                sys.exit(1)

    try:
        cmd, confirmed = input_loop()
    except KeyboardInterrupt:
        feedback({"event": "cancelled"})
        sys.exit(1)

    if cmd.script != confirmed or len(questions) > 1:
        # the user corrected the script
        db = corrections.get_db()
        q = "\n".join(questions)
        vector = similarity.embedding(text=q)
        db.add(vector, corrections.QATuple(question=q, answer=cmd.script, corrected=confirmed).to_dict())
        logging.debug("added correction to db")
        db.save(os.path.expanduser(settings.corrections_db_path))
        feedback({"event": "corrected"})

    feedback({"event": "run"})
    revised_cmd = copy(cmd)
    revised_cmd.script = confirmed
    rc, stdout = run(revised_cmd)
    feedback({"event": "finished", "rc": rc})

    while rc != 0:
        corrected_cmd = correction_loop("\n".join(questions), revised_cmd.script, stdout)
        if not corrected_cmd:
            return
        rc, stdout = run(corrected_cmd)
        if rc == 0:
            db = corrections.get_db()
            q = "\n".join(questions)
            vector = similarity.embedding(text=q)
            db.add(vector, corrections.QATuple(question=q, answer=cmd.script, corrected=corrected_cmd.script).to_dict())
            logging.debug("added correction to db")
            db.save(os.path.expanduser(settings.corrections_db_path))
        feedback({"event": "finished", "rc": rc, "corrected": True})


def query_mode(args):
    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO, format="%(message)s")
    settings.model = args.mode
    settings.history_context = args.history_context
    settings.request_explanation = args.explain
    try:
        do_query(" ".join(args.question))
        if settings_store.is_changed():
            for _ in range(3):
                answer = input("save the current settings? [y/n]")
                if answer == "y":
                    settings_store.save()
                    break
                elif answer == "n":
                    break
    except KeyboardInterrupt:
        pass


def query(
    *,
    q: str,
    prev_questions: list[str],
    prev_results: list[list[str]],
    corrected_results: list[corrections.QATuple],
    history_context: list[str],
    similar_history: list[str],
) -> list[Result]:
    req = {
        "query": q,
        "prev_questions": prev_questions,
        "prev_suggestions": prev_results,
        "context": {
            "history": history_context,
            "similar": similar_history,
            "corrections": [
                (qa.question, scrub_secrets.scrub_line("~/.bash_history", qa.corrected)) for qa in corrected_results
            ],
        },
        "model": settings.model,
        "trace_id": trace.trace_id,
        "platform": sys.platform,
        "shell": os.environ.get("SHELL"),
    }
    logging.debug("query with request: %s", req)
    response = requests.post(
        settings.endpoint + "/query",
        json=req,
        headers={
            "user-agent": f"promptops-cli; user_id={user.user_id()}",
        },
    )
    if response.status_code != 200:
        # this exception completely destroys the ui
        return []
        # raise Exception(f"there was problem with the response, status: {response.status_code}, text: {response.text}")

    data = response.json()
    try:
        return [Result.from_dict(entry) for entry in data["suggestions"]]
    except KeyError:
        logging.debug("response: %s", data)
        raise



from prompt_toolkit import prompt
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.formatted_text import HTML

kb = KeyBindings()


EXIT = "~ctrl_c"
TOGGLE_CLARIFY_MODE = "~toggle_clarify_mode"
GO_BACK = "~go_back"


@kb.add('c-c')
def _(event):
    event.app.exit(result=EXIT)


@kb.add('escape', eager=True)
def _(event):
    event.app.exit(result=GO_BACK)


@kb.add('c-r')
def _(event):
    event.app.exit(result=TOGGLE_CLARIFY_MODE)


def display_prompt(prompt_text, show_go_back=True):
    message = f"> "
    bottom_toolbar = HTML("<b>[enter]</b> confirm <b>[ctrl+c]</b> exit <b>[ctrl+r]</b> revise" + (" <b>[escape]</b> go back" if show_go_back else ""))
    user_input = prompt(message, default=prompt_text, bottom_toolbar=bottom_toolbar, key_bindings=kb)
    return user_input


def display_clarity_prompt(prompt_text):
    message = f"clarify: "
    bottom_toolbar = HTML("<b>[enter]</b> confirm <b>[ctrl+c]</b> exit <b>[escape]</b> go back")
    user_input = prompt(message, default=prompt_text, bottom_toolbar=bottom_toolbar, key_bindings=kb)
    return user_input



from prompt_toolkit import prompt
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.formatted_text import HTML
from prompt_toolkit.completion import Completer, Completion
from prompt_toolkit.shortcuts import CompleteStyle
import shlex
from pathlib import Path


kb = KeyBindings()


EXIT = "~ctrl_c"
TOGGLE_CLARIFY_MODE = "~toggle_clarify_mode"
SKIP = TOGGLE_CLARIFY_MODE
GO_BACK = "~go_back"


@kb.add("c-c")
def _(event):
    event.app.exit(result=EXIT)


@kb.add("escape", eager=True)
def _(event):
    event.app.exit(result=GO_BACK)


@kb.add("c-b", eager=True)
def _(event):
    event.app.exit(result=GO_BACK)


@kb.add("c-r")
def _(event):
    event.app.exit(result=TOGGLE_CLARIFY_MODE)


class FileCompleter(Completer):
    def get_completions(self, document, complete_event):
        # TODO: use shlex.shlex to get the lexer for more accurate word boundary
        shlex.shlex()
        adjust = 0
        try:
            words = shlex.split(document.text)
            if document.text == "" or document.text[-1] == " ":
                words.append("")
        except ValueError:
            adjust = -1
            try:
                words = shlex.split(document.text + "'")
            except ValueError:
                words = shlex.split(document.text + '"')
        word_before_cursor = words[-1]
        path = Path(word_before_cursor).expanduser()

        if path.is_dir():
            for file in sorted(path.iterdir(), key=lambda x: str(x)):
                yield Completion(
                    shlex.quote(str(file)), start_position=-len(word_before_cursor) + adjust, display=str(file)
                )
        else:
            for file in sorted(path.parent.iterdir(), key=lambda x: str(x)):
                if file.name.startswith(path.name):
                    yield Completion(
                        shlex.quote(str(file)),
                        start_position=-len(word_before_cursor) + adjust,
                        display=str(file),
                    )


def confirm_command(prompt_text, show_go_back=True, message="> "):
    bottom_toolbar = HTML(
        "<b>[enter]</b> confirm <b>[ctrl+c]</b> exit"
        + (" <b>[esc]</b> or <b>[ctrl+b]</b> go back" if show_go_back else "")
    )
    user_input = prompt(
        message,
        default=prompt_text,
        bottom_toolbar=bottom_toolbar,
        key_bindings=kb,
        completer=FileCompleter(),
        complete_style=CompleteStyle.READLINE_LIKE,
    )
    return user_input


def confirm(prompt_text):
    message = f"> "
    bottom_toolbar = HTML("<b>[enter]</b> confirm <b>[esc]</b> or <b>[ctrl+b]</b> go back")
    user_input = prompt(message, default=prompt_text, bottom_toolbar=bottom_toolbar, key_bindings=kb)
    return user_input


def confirm_clarify(prompt_text):
    message = f"clarify: "
    bottom_toolbar = HTML("<b>[enter]</b> confirm <b>[ctrl+c]</b> exit <b>[escape]</b> or <b>[ctrl+b]</b> go back")
    user_input = prompt(message, default=prompt_text, bottom_toolbar=bottom_toolbar, key_bindings=kb)
    return user_input



from concurrent import futures


class ThreadPoolExecutor(futures.ThreadPoolExecutor):
    def __init__(self, max_workers=None, thread_name_prefix="", initializer=None, initargs=()):
        super().__init__(max_workers, thread_name_prefix, initializer, initargs)
        self._futures = []
        self._active = True

    def submit(self, fn, *args, **kwargs):
        future = super().submit(fn, *args, **kwargs)
        self._futures.append(future)
        return future

    def shutdown(self, wait: bool = ..., *, cancel_futures: bool = ...) -> None:
        self._active = False
        for future in self._futures:
            future.cancel()
        super().shutdown(wait, cancel_futures=cancel_futures)

    def cancellable_callback(self, fn):
        def wrapper(*args, **kwargs):
            if self._active:
                return fn(*args, **kwargs)

        return wrapper

    def is_active(self):
        return self._active
